from django import forms
from django.contrib import messages
from django.db import models
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, FormView, ListView
from django.forms import formset_factory, BaseFormSet, HiddenInput

from byro.bookkeeping.models import Account, AccountCategory, Transaction
from byro.bookkeeping.models.transaction import TransactionManager
from byro.office.views.transactions import NewBookingForm

FORM_CLASS = forms.modelform_factory(Account, fields=["name", "account_category"])

ACCOUNT_COLUMN_HEADERS = {
    # FIXME Check this with an accountant who is a native english speaker
    AccountCategory.INCOME: (_("Charge"), _("Revenue")),
    AccountCategory.ASSET: (_("Increase"), _("Decrease")),
    AccountCategory.EQUITY: (_("Decrease"), _("Increase")),
    AccountCategory.LIABILITY: (_("Decrease"), _("Increase")),
    AccountCategory.EXPENSE: (_("Expense"), _("Rebate")),
}

class EnhancedBookingFormSet(BaseFormSet):
    def add_fields(self, form, index):
        super().add_fields(form, index)
        form.fields["transaction"] = forms.IntegerField(widget=HiddenInput)


class AccountListView(ListView):
    template_name = "office/account/list.html"
    context_object_name = "accounts"
    model = Account


class AccountCreateView(FormView):
    template_name = "office/account/add.html"
    model = Account
    form_class = FORM_CLASS

    def form_valid(self, form):
        form.save()
        messages.success(
            self.request,
            _("The account was added, please edit additional details if applicable."),
        )
        self.form = form
        return super().form_valid(form)

    def get_success_url(self):
        return reverse(
            "office:finance.accounts.detail", kwargs={"pk": self.form.instance.pk}
        )


class AccountDetailView(ListView):
    template_name = "office/account/detail.html"
    context_object_name = "bookings"
    model = Transaction
    paginate_by = 25

    def get_object(self):
        if not hasattr(self, "object"):
            self.object = Account.objects.get(pk=self.kwargs["pk"])
        return self.object

    def get_queryset(self):
        qs = self.get_object().bookings_with_transaction_data
        if self.request.GET.get("filter") == "unbalanced":
            qs = qs.exclude(
                transaction_balances_debit=models.F("transaction_balances_credit")
            )
        qs = qs.filter(transaction__value_datetime__lte=now()).order_by(
            "-transaction__value_datetime"
        )
        return qs

    def get_form(self, request=None):
        form = FORM_CLASS(request.POST if request else None, instance=self.get_object())
        form.fields["account_category"].disabled = True
        return form

    def get_formset(self, request=None):
        if not request:
            BookingFormSet = formset_factory(NewBookingForm, extra = len(self.get_queryset()), formset=EnhancedBookingFormSet)
            formset = BookingFormSet()
        else:
            BookingFormSet = formset_factory(NewBookingForm, extra = 0, formset=EnhancedBookingFormSet)
            formset = BookingFormSet(request.POST)
        return formset

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        formset = self.get_formset()
        context["formset"] = formset
        for index,booking in enumerate(context['bookings']):
            if booking.transaction_balances_credit > booking.transaction_balances_debit:
                formset[index].fields["debit_value"].initial = (
                    booking.transaction_balances_credit - booking.transaction_balances_debit
                )
            else:
                formset[index].fields["credit_value"].initial = (
                    booking.transaction_balances_debit - booking.transaction_balances_credit
                )
            booking.form = formset[index]
            booking.form.fields["transaction"].initial = booking.transaction.pk
        context["form"] = self.get_form()
        context["account"] = self.get_object()
        context["ACCOUNT_COLUMN_HEADERS"] = ACCOUNT_COLUMN_HEADERS.get(
            self.get_object().account_category, (_("Debit"), _("Credit"))
        )
        return context

    def post(self, request, *args, **kwargs):
        formset = self.get_formset(request)
        if formset.is_valid() and formset.has_changed():
            for form in formset:
                transaction_id = form.cleaned_data["transaction"]
                transaction = Transaction.objects.with_balances().get(pk=transaction_id)
                arguments = dict(memo = form.cleaned_data["memo"], account=form.cleaned_data["account"], member=form.cleaned_data["member"], importer="_manual_batch_entry", user_or_context=self)
                if form.cleaned_data["debit_value"]:
                    transaction.debit(amount=form.cleaned_data["debit_value"], **arguments)
                if form.cleaned_data["credit_value"]:
                    transaction.credit(amount=form.cleaned_data["credit_value"], **arguments)
                transaction.save()
                transaction.log(self, ".updated")
            messages.success(self.request, _("Your changes have been saved."))

        account = self.get_object()
        if account.unbalanced_transactions.count():
            return redirect("{}?filter=unbalanced".format(reverse("office:finance.accounts.detail",kwargs={"pk": account.pk},)))
        return redirect("office:finance.accounts.list")

class AccountDeleteView(DetailView):
    model = Account
    context_object_name = "account"
